package sun.nio.ch;

import java.util.HashMap;
import java.nio.channels.Selector;
import java.nio.channels.CancelledKeyException;
import java.nio.channels.SelectableChannel;
import java.util.Iterator;
import java.nio.channels.ClosedSelectorException;
import java.io.IOException;
import java.util.ArrayList;
import java.nio.channels.spi.SelectorProvider;
import java.nio.channels.Pipe;
import java.util.List;

final class WindowsSelectorImpl extends SelectorImpl
{
    private final int INIT_CAP = 8;
    private static final int MAX_SELECTABLE_FDS = 1024;
    private SelectionKeyImpl[] channelArray;
    private PollArrayWrapper pollWrapper;
    private int totalChannels;
    private int threadsCount;
    private final List<WindowsSelectorImpl.SelectThread> threads;
    private final Pipe wakeupPipe;
    private final int wakeupSourceFd;
    private final int wakeupSinkFd;
    private Object closeLock;
    private final WindowsSelectorImpl.FdMap fdMap;
    private final WindowsSelectorImpl.SubSelector subSelector;
    private long timeout;
    private final Object interruptLock;
    private volatile boolean interruptTriggered;
    private final WindowsSelectorImpl.StartLock startLock;
    private final WindowsSelectorImpl.FinishLock finishLock;
    private long updateCount;
    
    WindowsSelectorImpl(SelectorProvider selectorProvider) throws IOException {
        super(selectorProvider);
        this.channelArray = new SelectionKeyImpl[8];
        this.totalChannels = 1;
        this.threadsCount = 0;
        this.threads = new ArrayList();
        this.closeLock = new Object();
        this.fdMap = new WindowsSelectorImpl.FdMap();
        this.subSelector = new WindowsSelectorImpl.SubSelector();
        this.interruptLock = new Object();
        this.interruptTriggered = false;
        this.startLock = new WindowsSelectorImpl.StartLock();
        this.finishLock = new WindowsSelectorImpl.FinishLock();
        this.updateCount = 0L;
        this.pollWrapper = new PollArrayWrapper(8);
        this.wakeupPipe = Pipe.open();
        this.wakeupSourceFd = ((SelChImpl)this.wakeupPipe.source()).getFDVal();
        SinkChannelImpl sinkChannelImpl = (SinkChannelImpl)this.wakeupPipe.sink();
        sinkChannelImpl.sc.socket().setTcpNoDelay(true);
        this.wakeupSinkFd = sinkChannelImpl.getFDVal();
        this.pollWrapper.addWakeupSocket(this.wakeupSourceFd, 0);
    }
    
    @Override
    protected int doSelect(long timeout) throws IOException {
        if (this.channelArray == null) {
            throw new ClosedSelectorException();
        }
        this.timeout = timeout;
        this.processDeregisterQueue();
        if (this.interruptTriggered) {
            this.resetWakeupSocket();
            return 0;
        }
        this.adjustThreadsCount();
        WindowsSelectorImpl.FinishLock.access$200(this.finishLock);
        WindowsSelectorImpl.StartLock.access$300(this.startLock);
        try {
            this.begin();
            try {
                WindowsSelectorImpl.SubSelector.access$400(this.subSelector);
            }
            catch (IOException ex) {
                WindowsSelectorImpl.FinishLock.access$500(this.finishLock, ex);
            }
            if (this.threads.size() > 0) {
                WindowsSelectorImpl.FinishLock.access$600(this.finishLock);
            }
        }
        finally {
            this.end();
        }
        WindowsSelectorImpl.FinishLock.access$700(this.finishLock);
        this.processDeregisterQueue();
        int updateSelectedKeys = this.updateSelectedKeys();
        this.resetWakeupSocket();
        return updateSelectedKeys;
    }
    
    private void adjustThreadsCount() {
        if (this.threadsCount > this.threads.size()) {
            for (int j = this.threads.size(); j < this.threadsCount; ++j) {
                WindowsSelectorImpl.SelectThread o = new WindowsSelectorImpl.SelectThread(j);
                this.threads.add(o);
                o.setDaemon(true);
                o.start();
            }
        }
        else if (this.threadsCount < this.threads.size()) {
            for (int i = this.threads.size() - 1; i >= this.threadsCount; --i) {
                ((WindowsSelectorImpl.SelectThread)this.threads.remove(i)).makeZombie();
            }
        }
    }
    
    private void setWakeupSocket() {
        this.setWakeupSocket0(this.wakeupSinkFd);
    }
    
    private native void setWakeupSocket0(int p0);
    
    private void resetWakeupSocket() {
        synchronized (this.interruptLock) {
            if (!this.interruptTriggered) {
                return;
            }
            this.resetWakeupSocket0(this.wakeupSourceFd);
            this.interruptTriggered = false;
        }
    }
    
    private native void resetWakeupSocket0(int p0);
    
    private native boolean discardUrgentData(int p0);
    
    private int updateSelectedKeys() {
        ++this.updateCount;
        int n = 0 + WindowsSelectorImpl.SubSelector.access$2900(this.subSelector, this.updateCount);
        Iterator<WindowsSelectorImpl.SelectThread> iterator = this.threads.iterator();
        while (iterator.hasNext()) {
            n += WindowsSelectorImpl.SubSelector.access$2900(((WindowsSelectorImpl.SelectThread)iterator.next()).subSelector, this.updateCount);
        }
        return n;
    }
    
    @Override
    protected void implClose() throws IOException {
        synchronized (this.closeLock) {
            if (this.channelArray != null && this.pollWrapper != null) {
                synchronized (this.interruptLock) {
                    this.interruptTriggered = true;
                }
                this.wakeupPipe.sink().close();
                this.wakeupPipe.source().close();
                for (int i = 1; i < this.totalChannels; ++i) {
                    if (i % 1024 != 0) {
                        this.deregister(this.channelArray[i]);
                        SelectableChannel channel = this.channelArray[i].channel();
                        if (!channel.isOpen() && !channel.isRegistered()) {
                            ((SelChImpl)channel).kill();
                        }
                    }
                }
                this.pollWrapper.free();
                this.pollWrapper = null;
                this.selectedKeys = null;
                this.channelArray = null;
                Iterator<WindowsSelectorImpl.SelectThread> iterator = this.threads.iterator();
                while (iterator.hasNext()) {
                    ((WindowsSelectorImpl.SelectThread)iterator.next()).makeZombie();
                }
                WindowsSelectorImpl.StartLock.access$300(this.startLock);
            }
        }
    }
    
    @Override
    protected void implRegister(SelectionKeyImpl e) {
        synchronized (this.closeLock) {
            if (this.pollWrapper == null) {
                throw new ClosedSelectorException();
            }
            this.growIfNeeded();
            (this.channelArray[this.totalChannels] = e).setIndex(this.totalChannels);
            WindowsSelectorImpl.FdMap.access$3000(this.fdMap, e);
            this.keys.add(e);
            this.pollWrapper.addEntry(this.totalChannels, e);
            ++this.totalChannels;
        }
    }
    
    private void growIfNeeded() {
        if (this.channelArray.length == this.totalChannels) {
            int n = this.totalChannels * 2;
            SelectionKeyImpl[] channelArray = new SelectionKeyImpl[n];
            System.arraycopy(this.channelArray, 1, channelArray, 1, this.totalChannels - 1);
            this.channelArray = channelArray;
            this.pollWrapper.grow(n);
        }
        if (this.totalChannels % 1024 == 0) {
            this.pollWrapper.addWakeupSocket(this.wakeupSourceFd, this.totalChannels);
            ++this.totalChannels;
            ++this.threadsCount;
        }
    }
    
    @Override
    protected void implDereg(SelectionKeyImpl selectionKeyImpl) throws IOException {
        int index = selectionKeyImpl.getIndex();
        assert index >= 0;
        synchronized (this.closeLock) {
            if (index != this.totalChannels - 1) {
                (this.channelArray[index] = this.channelArray[this.totalChannels - 1]).setIndex(index);
                this.pollWrapper.replaceEntry(this.pollWrapper, this.totalChannels - 1, this.pollWrapper, index);
            }
            selectionKeyImpl.setIndex(-1);
        }
        this.channelArray[this.totalChannels - 1] = null;
        --this.totalChannels;
        if (this.totalChannels != 1 && this.totalChannels % 1024 == 1) {
            --this.totalChannels;
            --this.threadsCount;
        }
        WindowsSelectorImpl.FdMap.access$3100(this.fdMap, selectionKeyImpl);
        this.keys.remove(selectionKeyImpl);
        this.selectedKeys.remove(selectionKeyImpl);
        this.deregister(selectionKeyImpl);
        SelectableChannel channel = selectionKeyImpl.channel();
        if (!channel.isOpen() && !channel.isRegistered()) {
            ((SelChImpl)channel).kill();
        }
        return;
    }
    
    @Override
    public void putEventOps(SelectionKeyImpl selectionKeyImpl, int n) {
        synchronized (this.closeLock) {
            if (this.pollWrapper == null) {
                throw new ClosedSelectorException();
            }
            int index = selectionKeyImpl.getIndex();
            if (index == -1) {
                throw new CancelledKeyException();
            }
            this.pollWrapper.putEventOps(index, n);
        }
    }
    
    @Override
    public Selector wakeup() {
        synchronized (this.interruptLock) {
            if (!this.interruptTriggered) {
                this.setWakeupSocket();
                this.interruptTriggered = true;
            }
        }
        return this;
    }
    
    static boolean access$2200(WindowsSelectorImpl windowsSelectorImpl, int n) {
        return windowsSelectorImpl.discardUrgentData(n);
    }
    
    static {
        IOUtil.load();
    }
    
    private static final class FdMap extends HashMap<Integer, WindowsSelectorImpl.MapEntry>
    {
        static final long serialVersionUID = 0L;
        
        private WindowsSelectorImpl.MapEntry get(int value) {
            return (WindowsSelectorImpl.MapEntry)super.get(new Integer(value));
        }
        
        private WindowsSelectorImpl.MapEntry put(SelectionKeyImpl selectionKeyImpl) {
            return (WindowsSelectorImpl.MapEntry)super.put(new Integer(selectionKeyImpl.channel.getFDVal()), new MapEntry(selectionKeyImpl));
        }
        
        private WindowsSelectorImpl.MapEntry remove(SelectionKeyImpl selectionKeyImpl) {
            Integer n = new Integer(selectionKeyImpl.channel.getFDVal());
            WindowsSelectorImpl.MapEntry mapEntry = (WindowsSelectorImpl.MapEntry)super.get(n);
            if (mapEntry != null && mapEntry.ski.channel == selectionKeyImpl.channel) {
                return (WindowsSelectorImpl.MapEntry)super.remove(n);
            }
            return null;
        }
        
        static WindowsSelectorImpl.MapEntry access$2100(WindowsSelectorImpl.FdMap fdMap, int n) {
            return fdMap.get(n);
        }
        
        static WindowsSelectorImpl.MapEntry access$3000(WindowsSelectorImpl.FdMap fdMap, SelectionKeyImpl selectionKeyImpl) {
            return fdMap.put(selectionKeyImpl);
        }
        
        static WindowsSelectorImpl.MapEntry access$3100(WindowsSelectorImpl.FdMap fdMap, SelectionKeyImpl selectionKeyImpl) {
            return fdMap.remove(selectionKeyImpl);
        }
    }
    
    private static final class MapEntry
    {
        SelectionKeyImpl ski;
        long updateCount;
        long clearedCount;
        
        MapEntry(SelectionKeyImpl ski) {
            this.updateCount = 0L;
            this.clearedCount = 0L;
            this.ski = ski;
        }
    }
    
    private final class StartLock
    {
        private long runsCounter;
        
        private synchronized void startThreads() {
            ++this.runsCounter;
            this.notifyAll();
        }
        
        private synchronized boolean waitForStart(WindowsSelectorImpl.SelectThread selectThread) {
            while (this.runsCounter == selectThread.lastRun) {
                try {
                    WindowsSelectorImpl.this.startLock.wait();
                }
                catch (InterruptedException ex) {
                    Thread.currentThread().interrupt();
                }
            }
            if (selectThread.isZombie()) {
                return true;
            }
            selectThread.lastRun = this.runsCounter;
            return false;
        }
        
        static void access$300(WindowsSelectorImpl.StartLock startLock) {
            startLock.startThreads();
        }
        
        static boolean access$2500(WindowsSelectorImpl.StartLock startLock, WindowsSelectorImpl.SelectThread selectThread) {
            return startLock.waitForStart(selectThread);
        }
    }
    
    private final class FinishLock
    {
        private int threadsToFinish;
        IOException exception;
        
        private FinishLock() {
            this.exception = null;
        }
        
        private void reset() {
            this.threadsToFinish = WindowsSelectorImpl.this.threads.size();
        }
        
        private synchronized void threadFinished() {
            if (this.threadsToFinish == WindowsSelectorImpl.this.threads.size()) {
                WindowsSelectorImpl.this.wakeup();
            }
            --this.threadsToFinish;
            if (this.threadsToFinish == 0) {
                this.notify();
            }
        }
        
        private synchronized void waitForHelperThreads() {
            if (this.threadsToFinish == WindowsSelectorImpl.this.threads.size()) {
                WindowsSelectorImpl.this.wakeup();
            }
            while (this.threadsToFinish != 0) {
                try {
                    WindowsSelectorImpl.this.finishLock.wait();
                }
                catch (InterruptedException ex) {
                    Thread.currentThread().interrupt();
                }
            }
        }
        
        private synchronized void setException(IOException exception) {
            this.exception = exception;
        }
        
        private void checkForException() throws IOException {
            if (this.exception == null) {
                return;
            }
            StringBuffer sb = new StringBuffer("An exception occurred during the execution of select(): \n");
            sb.append(this.exception);
            sb.append('\n');
            this.exception = null;
            throw new IOException(sb.toString());
        }
        
        static void access$200(WindowsSelectorImpl.FinishLock finishLock) {
            finishLock.reset();
        }
        
        static void access$500(WindowsSelectorImpl.FinishLock finishLock, IOException exception) {
            finishLock.setException(exception);
        }
        
        static void access$600(WindowsSelectorImpl.FinishLock finishLock) {
            finishLock.waitForHelperThreads();
        }
        
        static void access$700(WindowsSelectorImpl.FinishLock finishLock) throws IOException {
            finishLock.checkForException();
        }
        
        static void access$2700(WindowsSelectorImpl.FinishLock finishLock) {
            finishLock.threadFinished();
        }
    }
    
    private final class SubSelector
    {
        private final int pollArrayIndex;
        private final int[] readFds;
        private final int[] writeFds;
        private final int[] exceptFds;
        
        private SubSelector() {
            this.readFds = new int[1025];
            this.writeFds = new int[1025];
            this.exceptFds = new int[1025];
            this.pollArrayIndex = 0;
        }
        
        private SubSelector(int n) {
            this.readFds = new int[1025];
            this.writeFds = new int[1025];
            this.exceptFds = new int[1025];
            this.pollArrayIndex = (n + 1) * 1024;
        }
        
        private int poll() throws IOException {
            return this.poll0(WindowsSelectorImpl.this.pollWrapper.pollArrayAddress, Math.min(WindowsSelectorImpl.this.totalChannels, 1024), this.readFds, this.writeFds, this.exceptFds, WindowsSelectorImpl.this.timeout);
        }
        
        private int poll(int n) throws IOException {
            return this.poll0(WindowsSelectorImpl.this.pollWrapper.pollArrayAddress + (long)(this.pollArrayIndex * PollArrayWrapper.SIZE_POLLFD), Math.min(1024, WindowsSelectorImpl.this.totalChannels - (n + 1) * 1024), this.readFds, this.writeFds, this.exceptFds, WindowsSelectorImpl.this.timeout);
        }
        
        private native int poll0(long p0, int p1, int[] p2, int[] p3, int[] p4, long p5);
        
        private int processSelectedKeys(long n) {
            return 0 + this.processFDSet(n, this.readFds, (int)Net.POLLIN, false) + this.processFDSet(n, this.writeFds, Net.POLLCONN | Net.POLLOUT, false) + this.processFDSet(n, this.exceptFds, Net.POLLIN | Net.POLLCONN | Net.POLLOUT, true);
        }
        
        private int processFDSet(long n, int[] array, int n2, boolean b) {
            int n3 = 0;
            for (int i = 1; i <= array[0]; ++i) {
                int n4 = array[i];
                if (n4 == WindowsSelectorImpl.this.wakeupSourceFd) {
                    synchronized (WindowsSelectorImpl.this.interruptLock) {
                        WindowsSelectorImpl.this.interruptTriggered = true;
                    }
                }
                else {
                    WindowsSelectorImpl.MapEntry access$2100 = WindowsSelectorImpl.FdMap.access$2100(WindowsSelectorImpl.this.fdMap, n4);
                    if (access$2100 != null) {
                        SelectionKeyImpl ski = access$2100.ski;
                        if (!b || !(ski.channel() instanceof SocketChannelImpl) || !WindowsSelectorImpl.access$2200(WindowsSelectorImpl.this, n4)) {
                            if (WindowsSelectorImpl.this.selectedKeys.contains(ski)) {
                                if (access$2100.clearedCount != n) {
                                    if (ski.channel.translateAndSetReadyOps(n2, ski) && access$2100.updateCount != n) {
                                        access$2100.updateCount = n;
                                        ++n3;
                                    }
                                }
                                else if (ski.channel.translateAndUpdateReadyOps(n2, ski) && access$2100.updateCount != n) {
                                    access$2100.updateCount = n;
                                    ++n3;
                                }
                                access$2100.clearedCount = n;
                            }
                            else {
                                if (access$2100.clearedCount != n) {
                                    ski.channel.translateAndSetReadyOps(n2, ski);
                                    if ((ski.nioReadyOps() & ski.nioInterestOps()) != 0x0) {
                                        WindowsSelectorImpl.this.selectedKeys.add(ski);
                                        access$2100.updateCount = n;
                                        ++n3;
                                    }
                                }
                                else {
                                    ski.channel.translateAndUpdateReadyOps(n2, ski);
                                    if ((ski.nioReadyOps() & ski.nioInterestOps()) != 0x0) {
                                        WindowsSelectorImpl.this.selectedKeys.add(ski);
                                        access$2100.updateCount = n;
                                        ++n3;
                                    }
                                }
                                access$2100.clearedCount = n;
                            }
                        }
                    }
                }
            }
            return n3;
        }
        
        static int access$400(WindowsSelectorImpl.SubSelector subSelector) throws IOException {
            return subSelector.poll();
        }
        
        static int access$2600(WindowsSelectorImpl.SubSelector subSelector, int n) throws IOException {
            return subSelector.poll(n);
        }
        
        static int access$2900(WindowsSelectorImpl.SubSelector subSelector, long n) {
            return subSelector.processSelectedKeys(n);
        }
    }
    
    private final class SelectThread extends Thread
    {
        private final int index;
        final WindowsSelectorImpl.SubSelector subSelector;
        private long lastRun;
        private volatile boolean zombie;
        
        private SelectThread(int index) {
            this.lastRun = 0L;
            this.index = index;
            this.subSelector = new WindowsSelectorImpl.SubSelector(index);
            this.lastRun = WindowsSelectorImpl.this.startLock.runsCounter;
        }
        
        void makeZombie() {
            this.zombie = true;
        }
        
        boolean isZombie() {
            return this.zombie;
        }
        
        @Override
        public void run() {
            while (!WindowsSelectorImpl.StartLock.access$2500(WindowsSelectorImpl.this.startLock, this)) {
                try {
                    WindowsSelectorImpl.SubSelector.access$2600(this.subSelector, this.index);
                }
                catch (IOException ex) {
                    WindowsSelectorImpl.FinishLock.access$500(WindowsSelectorImpl.this.finishLock, ex);
                }
                WindowsSelectorImpl.FinishLock.access$2700(WindowsSelectorImpl.this.finishLock);
            }
        }
    }
}
